@@ -14,9 +14,9 @@ use Graph::AdjacencyMap qw(:flags :fields);
 
 use vars qw($VERSION);
 
-$VERSION = '0.69';
+$VERSION = '0.78';
 
-require 5.006; # Weak references are required.
+require 5.006; # Weak references are absolutely required.
 
 use Graph::AdjacencyMap::Heavy;
 use Graph::AdjacencyMap::Light;
@@ -349,7 +349,18 @@ sub add_vertex {
 	    Carp::croak("Graph::add_vertex: use add_vertices for more than one vertex or use hypervertexed");
 	}
 	for my $v ( @_ ) {
-	    $g->[ _V ]->set_path( $v ) unless $g->has_vertex( $v );
+	    if (defined $v) {
+		$g->[ _V ]->set_path( $v ) unless $g->has_vertex( $v );
+	    } else {
+		require Carp;
+		Carp::croak("Graph::add_vertex: undef vertex");
+	    }
+	}
+    }
+    for my $v ( @_ ) {
+	unless (defined $v) {
+	    require Carp;
+	    Carp::croak("Graph::add_vertex: undef vertex");
 	}
     }
     $g->[ _V ]->set_path( @_ );
@@ -429,8 +440,9 @@ sub _add_edge {
 	    push @e, $V->[ _s ]->{ $v };
 	}
     } else {
+	my $h = $g->[ _V ]->_is_HYPER;
 	for my $v ( @_ ) {
-	    my @v = ref $v eq 'ARRAY' ? @$v : $v;
+	    my @v = ref $v eq 'ARRAY' && $h ? @$v : $v;
 	    $g->add_vertex( @v ) unless $V->has_path( @v );
 	    push @e, $V->_get_path_id( @v );
 	}
@@ -475,8 +487,9 @@ sub _vertex_ids {
 	    push @e, $V->[ _s ]->{ $v };
 	}
     } else {
+	my $h = $g->[ _V ]->_is_HYPER;
 	for my $v ( @_ ) {
-	    my @v = ref $v eq 'ARRAY' ? @$v : $v;
+	    my @v = ref $v eq 'ARRAY' && $h ? @$v : $v;
 	    return () unless $V->has_path( @v );
 	    push @e, $V->_get_path_id( @v );
 	}
@@ -690,8 +703,9 @@ sub vertices_at {
     my @v;
     while (my ($i, $v) = each %{ $Vi }) {
 	my %i;
+	my $h = $V->[_f ] & _HYPER;
 	@i{ @i } = @i if @i; # @todo: nonuniq hyper vertices?
-	for my $u ((ref($v) eq 'ARRAY') ? @$v : $v) {
+	for my $u (ref $v eq 'ARRAY' && $h ? @$v : $v) {
 	    my $j = exists $v{ $u } ? $v{ $u } : ( $v{ $u } = $i );
 	    if (defined $j && exists $i{ $j }) {
 		delete $i{ $j };
@@ -712,10 +726,9 @@ sub _edges_at {
     my @e;
     my $en = 0;
     my %ev;
-    for my $v ( ($V->[_f ] & _HYPER) ? $g->vertices_at( @_ ) : @_ ) {
-#use Carp;
-#confess("v: $v, ref=".ref($v)) unless ref($v) eq 'ARRAY'; # Added by bsb
-	my $vi = $V->_get_path_id( (ref($v) eq 'ARRAY') ? @$v : $v );
+    my $h = $V->[_f ] & _HYPER;
+    for my $v ( $h ? $g->vertices_at( @_ ) : @_ ) {
+	my $vi = $V->_get_path_id( ref $v eq 'ARRAY' && $h ? @$v : $v );
 	next unless defined $vi;
 	my $Ei = $E->_ids;
 	while (my ($ei, $ev) = each %{ $Ei }) {
@@ -742,8 +755,9 @@ sub _edges_from {
     my $o = $E->[ _f ] & _UNORD;
     my $en = 0;
     my %ev;
-    for my $v ( ($V->[_f ] & _HYPER) ? $g->vertices_at( @_ ) : @_ ) {
-	my $vi = $V->_get_path_id( (ref($v) eq 'ARRAY') ? @$v : $v );
+    my $h = $V->[_f ] & _HYPER;
+    for my $v ( $h ? $g->vertices_at( @_ ) : @_ ) {
+	my $vi = $V->_get_path_id( ref $v eq 'ARRAY' && $h ? @$v : $v );
 	next unless defined $vi;
 	my $Ei = $E->_ids;
 	if (wantarray) {
@@ -793,8 +807,9 @@ sub _edges_to {
     my $o = $E->[ _f ] & _UNORD;
     my $en = 0;
     my %ev;
-    for my $v ( ($V->[_f ] & _HYPER) ? $g->vertices_at( @_ ) : @_ ) {
-	my $vi = $V->_get_path_id( (ref($v) eq 'ARRAY') ? @$v : $v );
+    my $h = $V->[_f ] & _HYPER;
+    for my $v ( $h ? $g->vertices_at( @_ ) : @_ ) {
+	my $vi = $V->_get_path_id( ref $v eq 'ARRAY' && $h ? @$v : $v );
 	next unless defined $vi;
 	my $Ei = $E->_ids;
 	if (wantarray) {
@@ -1165,7 +1180,7 @@ sub delete_cycle {
 
 sub has_cycle {
     my $g = shift;
-    $g->has_path(@_, $_[0]);
+    @_ ? ($g->has_path(@_, $_[0]) ? 1 : 0) : 0;
 }
 
 sub has_a_cycle {
@@ -1474,7 +1489,7 @@ sub get_edge_attribute {
     my $g = shift;
     $g->expect_non_multiedged;
     my $attr = pop;
-    return unless $g->has_edge( @_ );
+    return undef unless $g->has_edge( @_ );
     my @i = $g->_vertex_ids( @_ );
     return undef if @i == 0 && @_;
     my $E = $g->[ _E ];
@@ -1563,14 +1578,14 @@ sub vertex {
 
 sub out_edges {
     my $g = shift;
-    return undef unless @_ && $g->has_vertex( @_ );
+    return unless @_ && $g->has_vertex( @_ );
     my @e = $g->edges_from( @_ );
     wantarray ? map { @$_ } @e : @e;
 }
 
 sub in_edges {
     my $g = shift;
-    return undef unless @_ && $g->has_vertex( @_ );
+    return unless @_ && $g->has_vertex( @_ );
     my @e = $g->edges_to( @_ );
     wantarray ? map { @$_ } @e : @e;
 }
@@ -1588,7 +1603,8 @@ sub add_edges {
 	    $g->add_edge( @$u );
 	} else {
 	    if (@_) {
-		$g->add_edge( $u, shift @_ );
+		my $v = shift @_;
+		$g->add_edge( $u, $v );
 	    } else {
 		require Carp;
 		Carp::croak("Graph::add_edges: missing end vertex");
@@ -2016,7 +2032,7 @@ sub _get_options {
 # Random constructors and accessors.
 #
 
-sub fisher_yates_shuffle (@) {
+sub __fisher_yates_shuffle (@) {
     # From perlfaq4, but modified to be non-modifying.
     my @a = @_;
     my $i = @a;
@@ -2037,7 +2053,7 @@ BEGIN {
     # bleadperl that calls itself 5.9.3 but doesn't yet have the
     # patches, oh, well.
     *_shuffle = $^P && $] < 5.009003 ?
-	\&fisher_yates_shuffle : \&List::Util::shuffle;
+	\&__fisher_yates_shuffle : \&List::Util::shuffle;
 }
 
 sub random_graph {
@@ -2234,16 +2250,18 @@ sub _root_opt {
 }
 
 sub _heap_walk {
-    my ($g, $h, $add, $etc) = splice @_, 0, 4;
+    my ($g, $h, $add, $etc) = splice @_, 0, 4; # Leave %opt in @_.
 
     my ($opt, $unseenh, $unseena, $r, $next, $code, $attr) = $g->_root_opt(@_);
     my $HF = Heap::Fibonacci->new;
 
     while (defined $r) {
+	# print "r = $r\n";
 	$add->($g, $h, $HF, $r, $attr, $unseenh, $etc);
 	delete $unseenh->{ $r };
 	while (defined $HF->top) {
 	    my $t = $HF->extract_top;
+	    # use Data::Dumper; print "t = ", Dumper($t);
 	    if (defined $t) {
 		my ($u, $v, $w) = $t->val;
 		# print "extracted top: $u $v $w\n";
@@ -2539,13 +2557,14 @@ sub _connected_components {
 		@cci{ keys %icci } = values %icci;
 	    }
 	} else {
-	    my %r; @r{ $g->unique_vertices } = ();
+	    my @u = $g->unique_vertices;
+	    my %r; @r{ @u } = @u;
 	    my $froot = sub {
-		wantarray ? ( keys %r ? scalar each %r: ()) : each %r;
+		(each %r)[1];
 	    };
 	    my $nroot = sub {
-		$cc++;
-		wantarray ? ( keys %r ? scalar each %r: ()) : each %r;
+		$cc++ if keys %r;
+		(each %r)[1];
 	    };
 	    my $t = Graph::Traversal::DFS->new($g,
 					       first_root => $froot,
@@ -2614,17 +2633,21 @@ sub same_connected_components {
     }
 }
 
-my $super_component = sub { join("+", sort @{ $_[0] }) };
+my $super_component = sub { join("+", sort @_) };
 
 sub connected_graph {
-    my $g = shift;
+    my ($g, %opt) = @_;
     $g->expect_undirected;
     my $cg = Graph->new(undirected => 1);
     if ($g->has_union_find && $g->vertices == 1) {
+	# TODO: super_component?
 	$cg->add_vertices($g->vertices);
     } else {
+	my $sc_cb =
+	    exists $opt{super_component} ?
+		$opt{super_component} : $super_component;
 	for my $cc ( $g->connected_components() ) {
-	    my $sc = $super_component->($cc);
+	    my $sc = $sc_cb->(@$cc);
 	    $cg->add_vertex($sc);
 	    $cg->set_vertex_attribute($sc, 'subvertices', [ @$cc ]);
 	}
@@ -2729,7 +2752,7 @@ sub strongly_connected_component_by_vertex {
 	    return $i if $scc[$i]->[$j] eq $v;
 	}
     }
-    return undef;
+    return;
 }
 
 sub strongly_connected_component_by_index {
@@ -2829,18 +2852,27 @@ sub strongly_connected_graph {
 
     $u->dfs;
 
-    my $hypervertex = $super_component;
+    my $sc_cb;
+    my $hv_cb;
 
-    _opt_get(\%attr, hypervertex => \$hypervertex);
+    _opt_get(\%attr, super_component => \$sc_cb);
+    _opt_get(\%attr, hypervertex => \$hv_cb);
     _opt_unknown(\%attr);
 
+    if (defined $hv_cb && !defined $sc_cb) {
+	$sc_cb = sub { $hv_cb->( [ @_ ] ) };
+    }
+    unless (defined $sc_cb) {
+	$sc_cb = $super_component;
+    }
+
     my $s = Graph->new;
 
     my %c;
     my @s;
     for (my $i = 0; $i <  @c; $i++) {
 	my $c = $c[$i];
-	$s->add_vertex( $s[$i] = $hypervertex->($c) );
+	$s->add_vertex( $s[$i] = $sc_cb->(@$c) );
 	$s->set_vertex_attribute($s[$i], 'subvertices', [ @$c ]);
 	for my $v (@$c) {
 	    $c{$v} = $i;
@@ -2880,10 +2912,10 @@ sub _make_bcc {
     my %b;
     while (@$S) {
 	my $t = pop @$S;
-	$b{ $t }++;
+	$b{ $t } = $t;
 	last if $t eq $v;
     }
-    return [ keys %b, $c ];
+    return [ values %b, $c ];
 }
 
 sub biconnectivity {
@@ -2914,84 +2946,140 @@ sub biconnectivity {
 	my %L = ( $r => 1 );
 	my @S = ( $r );
 	my %A;
-	for my $v ($g->vertices) {
-	    @{ $A{ $v } }{ $g->successors( $v ) } = ();
+	my @V = $g->vertices;
+
+	# print "V : @V\n";
+	# print "r : $r\n";
+
+	my %T; @T{ @V } = @V;
+	my %S; # Self-loops.
+
+	for my $w (@V) {
+	    my @s = $g->successors( $w );
+	    if (@s) {
+		@s = grep { $_ eq $w ? ( delete $T{ $w }, 0 ) : 1 } @s;
+		@{ $A{ $w } }{ @s } = @s;
+	    } elsif ($g->predecessors( $w ) == 0) {
+		delete $T{ $w };
+		if ($w eq $r) {
+		    delete $I { $r };
+		    $r = $v = each %T;
+		    if (defined $r) {
+			%L = ( $r => 1 );
+			@S = ( $r );
+			$I{ $r } = 1;
+			# print "r : $r\n";
+		    }
+		}
+	    }
 	}
-	my @C;
-	my $Avok;
-	do {
-	    my $w;
-	    do {
-		$w = first { !$U{ $v }{ $_ } } _shuffle keys %{ $A{ $v } };
-		if (defined $w) {
-		    $U{ $v }{ $w }++;
-		    $U{ $w }{ $v }++;
-		    if ($I{ $w } == 0) {
-			$P{ $w } = $v;
-			$i++;
-			$I{ $w } = $i;
-			$L{ $w } = $i;
-			push @S, $w;
-			$v = $w;
-		    } else {
-			$L{ $v } = $I{ $w } if $I{ $w } < $L{ $v };
-		    }
-		}
-	    } while (defined $w);
-	    if (!defined $P{ $v }) {
-		# Do nothing.
-	    } elsif ($P{ $v } ne $r) {
-		if ($L{ $v } < $I{ $P{ $v } }) {
-		    $L{ $P{ $v } } = $L{ $v } if $L{ $v } < $L{ $P{ $v } };
-		} else {
-		    $AP{ $P{ $v } }++;
-		    push @C, _make_bcc(\@S, $v, $P{ $v } );
-		}
-	    } else {
-		my $e;
-		for my $w (_shuffle keys %{ $A{ $r } }) {
-		    unless ($U{ $r }{ $w }) {
-			$e = $r;
-			last;
-		    }
-		}
-		$AP{ $e }++ if defined $e;
-		push @C, _make_bcc(\@S, $v, $r);
-	    }
-	    $v = defined $P{ $v } ? $P{ $v } : $r;
-	    $Avok = 0;
-	    if (defined $v) {
-		$Avok = 0;
-		if (keys %{ $A{ $v } }) {
-		    if (!exists $P{ $v }) {
-			for my $w (keys %{ $A{ $v } }) {
-			    $Avok++ if $U{ $v }{ $w };
-			}
-			$Avok = 0 unless $Avok == keys %{ $A{ $v } };
-		    }
-		} else {
-		    $Avok = 1;
-		}
-	    }
-	} until ($Avok);
+
+	# use Data::Dumper;
+	# print "T : ", Dumper(\%T);
+	# print "A : ", Dumper(\%A);
 
 	my %V2BC;
-	for (my $i = 0; $i < @C; $i++) {
-	    for my $v (@{ $C[ $i ]}) {
-		$V2BC{ $v }{ $i }++;
+	my @BR;
+	my @BC;
+
+	my @C;
+	my $Avok;
+
+	while (keys %T) {
+	    # print "T = ", Dumper(\%T);
+	    do {
+		my $w;
+	        do {
+		    my @w = _shuffle values %{ $A{ $v } };
+		    # print "w = @w\n";
+		    $w = first { !$U{ $v }{ $_ } } @w;
+		    if (defined $w) {
+			# print "w = $w\n";
+			$U{ $v }{ $w }++;
+			$U{ $w }{ $v }++;
+			if ($I{ $w } == 0) {
+			    $P{ $w } = $v;
+			    $i++;
+			    $I{ $w } = $i;
+			    $L{ $w } = $i;
+			    push @S, $w;
+			    $v = $w;
+			} else {
+			    $L{ $v } = $I{ $w } if $I{ $w } < $L{ $v };
+			}
+		    }
+		} while (defined $w);
+		# print "U = ", Dumper(\%U);
+		# print "P = ", Dumper(\%P);
+		# print "L = ", Dumper(\%L);
+		if (!defined $P{ $v }) {
+		    # Do nothing.
+		} elsif ($P{ $v } ne $r) {
+		    if ($L{ $v } < $I{ $P{ $v } }) {
+			$L{ $P{ $v } } = $L{ $v } if $L{ $v } < $L{ $P{ $v } };
+		    } else {
+			$AP{ $P{ $v } } = $P{ $v };
+			push @C, _make_bcc(\@S, $v, $P{ $v } );
+		    }
+		} else {
+		    my $e;
+		    for my $w (_shuffle keys %{ $A{ $r } }) {
+			# print "w = $w\n";
+			unless ($U{ $r }{ $w }) {
+			    $e = $r;
+			    # print "e = $e\n";
+			    last;
+			}
+		    }
+		    $AP{ $e } = $e if defined $e;
+		    push @C, _make_bcc(\@S, $v, $r);
+		}
+		# print "AP = ", Dumper(\%AP);
+		# print "C  = ", Dumper(\@C);
+		# print "L  = ", Dumper(\%L);
+		$v = defined $P{ $v } ? $P{ $v } : $r;
+		# print "v = $v\n";
+		$Avok = 0;
+		if (defined $v) {
+		    if (keys %{ $A{ $v } }) {
+			if (!exists $P{ $v }) {
+			    for my $w (keys %{ $A{ $v } }) {
+				$Avok++ if $U{ $v }{ $w };
+			    }
+			    # print "Avok/1 = $Avok\n";
+			    $Avok = 0 unless $Avok == keys %{ $A{ $v } };
+			    # print "Avok/2 = $Avok\n";
+			}
+		    } else {
+			$Avok = 1;
+			# print "Avok/3 = $Avok\n";
+		    }
+		}
+	    } until ($Avok);
+
+	    last if @C == 0 && !exists $S{$v};
+
+	    for (my $i = 0; $i < @C; $i++) {
+		for my $v (@{ $C[ $i ]}) {
+		    $V2BC{ $v }{ $i }++;
+		    delete $T{ $v };
+		}
+	    }
+
+	    for (my $i = 0; $i < @C; $i++) {
+		if (@{ $C[ $i ] } == 2) {
+		    push @BR, $C[ $i ];
+		} else {
+		    push @BC, $C[ $i ];
+		}
+	    }
+
+	    if (keys %T) {
+		$r = $v = each %T;
 	    }
 	}
 
-	my @BR;
-	my @BC;
-	for (my $i = 0; $i < @C; $i++) {
-	    if (@{ $C[ $i ] } == 2) {
-		push @BR, $C[ $i ];
-	    } else {
-		push @BC, $C[ $i ];
-	    }
-	}
-	$bcc = [ [keys %AP], \@BC, \@BR, \%V2BC ];
+	$bcc = [ [values %AP], \@BC, \@BR, \%V2BC ];
 	$g->set_graph_attribute('_bcc', [ $g->[ _G ], $bcc ]);
     }
 
@@ -3067,11 +3155,14 @@ sub same_biconnected_components {
 }
 
 sub biconnected_graph {
-    my $g = shift;
+    my ($g, %opt) = @_;
     my ($bc, $v2bc) = ($g->biconnectivity)[1, 3];
     my $bcg = Graph::Undirected->new;
+    my $sc_cb =
+	exists $opt{super_component} ?
+	    $opt{super_component} : $super_component;
     for my $c (@$bc) {
-	$bcg->add_vertex(my $s = $super_component->($c));
+	$bcg->add_vertex(my $s = $sc_cb->(@$c));
 	$bcg->set_vertex_attribute($s, 'subvertices', [ @$c ]);
     }
     my %k;
@@ -3085,8 +3176,8 @@ sub biconnected_graph {
 		for my $u (@u) {
 		    if (exists $j{ $u }) {
 			unless ($k{ $i }{ $j }++) {
-			    $bcg->add_edge($super_component->($bc->[$i]),
-					   $super_component->($bc->[$j]));
+			    $bcg->add_edge($sc_cb->(@{$bc->[$i]}),
+					   $sc_cb->(@{$bc->[$j]}));
 			}
 			last;
 		    }
@@ -3109,18 +3200,21 @@ sub bridges {
 
 sub _SPT_add {
     my ($g, $h, $HF, $r, $attr, $unseen, $etc) = @_;
+    my $etc_r = $etc->{ $r } || 0;
     for my $s ( grep { exists $unseen->{ $_ } } $g->successors( $r ) ) {
-	my $t = $g->get_edge_attribute( $r, $s, $attr ) || 1;
+	my $t = $g->get_edge_attribute( $r, $s, $attr );
+	$t = 1 unless defined $t;
 	if ($t < 0) {
 	    require Carp;
 	    Carp::croak("Graph::SPT_Dijkstra: edge $r-$s is negative ($t)");
 	}
-	if (!defined($etc->{ $s }) || $etc->{ $r } + $t < $etc->{ $s }) {
-	    $etc->{ $s } = ($etc->{ $r } || 0) + $t;
-	    # print "$r - $s : setting $s to $etc->{ $s }\n";
-	    $h->set_vertex_attribute( $s, $attr, $etc->{ $s } );
+	if (!defined($etc->{ $s }) || ($etc_r + $t) < $etc->{ $s }) {
+	    my $etc_s = $etc->{ $s } || 0;
+	    $etc->{ $s } = $etc_r + $t;
+	    # print "$r - $s : setting $s to $etc->{ $s } ($etc_r, $etc_s)\n";
+	    $h->set_vertex_attribute( $s, $attr, $etc->{ $s });
 	    $h->set_vertex_attribute( $s, 'p', $r );
-	    $HF->add( Graph::SPTHeapElem->new($r, $s, $t) );
+	    $HF->add( Graph::SPTHeapElem->new($r, $s, $etc->{ $s }) );
 	}
     }
 }
@@ -3130,16 +3224,21 @@ sub SPT_Dijkstra {
     my %opt = @_ == 1 ? (first_root => $_[0]) : @_;
     my $first_root = $opt{ first_root };
     unless (defined $first_root) {
-	$opt{ first_root} = $first_root = $g->random_vertex();
+	$opt{ first_root } = $first_root = $g->random_vertex();
     }
     my $spt_di = $g->get_graph_attribute('_spt_di');
     unless (defined $spt_di && exists $spt_di->{ $first_root } && $spt_di->{ $first_root }->[ 0 ] == $g->[ _G ]) {
 	my %etc;
-	my $sptg = $g->_heap_walk($g->new(), \&_SPT_add, \%etc, %opt);
+	my $sptg = $g->_heap_walk($g->new, \&_SPT_add, \%etc, %opt);
 	$spt_di->{ $first_root } = [ $g->[ _G ], $sptg ];
 	$g->set_graph_attribute('_spt_di', $spt_di);
     }
-    return $spt_di->{ $first_root }->[ 1 ];
+
+    my $spt = $spt_di->{ $first_root }->[ 1 ];
+
+    $spt->set_graph_attribute('SPT_Dijsktra_root', $first_root);
+
+    return $spt;
 }
 
 *SSSP_Dijkstra = \&SPT_Dijkstra;
@@ -3150,19 +3249,47 @@ sub SP_Dijkstra {
     my ($g, $u, $v) = @_;
     my $sptg = $g->SPT_Dijkstra(first_root => $u);
     my @path = ($v);
+    my %seen;
+    my $V = $g->vertices;
     my $p;
     while (defined($p = $sptg->get_vertex_attribute($v, 'p'))) {
+	last if exists $seen{$p};
 	push @path, $p;
 	$v = $p;
+	$seen{$p}++;
+	last if keys %seen == $V || $u eq $v;
     }
     @path = () if @path && $path[-1] ne $u;
     return reverse @path;
 }
 
+sub __SPT_Bellman_Ford {
+    my ($g, $u, $v, $attr, $d, $p, $c0, $c1) = @_;
+    return unless $c0->{ $u };
+    my $w = $g->get_edge_attribute($u, $v, $attr);
+    $w = 1 unless defined $w;
+    if (defined $d->{ $v }) {
+	if (defined $d->{ $u }) {
+	    if ($d->{ $v } > $d->{ $u } + $w) {
+		$d->{ $v } = $d->{ $u } + $w;
+		$p->{ $v } = $u;
+		$c1->{ $v }++;
+	    }
+	} # else !defined $d->{ $u } &&  defined $d->{ $v }
+    } else {
+	if (defined $d->{ $u }) {
+	    #  defined $d->{ $u } && !defined $d->{ $v }
+	    $d->{ $v } = $d->{ $u } + $w;
+	    $p->{ $v } = $u;
+	    $c1->{ $v }++;
+	} # else !defined $d->{ $u } && !defined $d->{ $v }
+    }
+}
+
 sub _SPT_Bellman_Ford {
     my ($g, $opt, $unseenh, $unseena, $r, $next, $code, $attr) = @_;
-
     my %d;
+    return unless defined $r;
     $d{ $r } = 0;
     my %p;
     my $V = $g->vertices;
@@ -3172,24 +3299,9 @@ sub _SPT_Bellman_Ford {
 	my %c1;
 	for my $e ($g->edges) {
 	    my ($u, $v) = @$e;
-	    next unless $c0{ $u };
-	    my $w = $g->get_edge_attribute($u, $v, $attr);
-	    $w = 1 unless defined $w;
-	    if (defined $d{ $v }) {
-		if (defined $d{ $u }) {
-		    if ($d{ $v } > $d{ $u } + $w) {
-			$d{ $v } = $d{ $u } + $w;
-			$p{ $v } = $u;
-			$c1{ $v }++;
-		    }
-		} # else !defined $d{ $u } &&  defined $d{ $v }
-	    } else {
-		if (defined $d{ $u }) {
-		    #  defined $d{ $u } && !defined $d{ $v }
-		    $d{ $v } = $d{ $u } + $w;
-		    $p{ $v } = $u;
-		    $c1{ $v }++;
-		} # else !defined $d{ $u } && !defined $d{ $v }
+	    __SPT_Bellman_Ford($g, $u, $v, $attr, \%d, \%p, \%c0, \%c1);
+	    if ($g->undirected) {
+		__SPT_Bellman_Ford($g, $v, $u, $attr, \%d, \%p, \%c0, \%c1);
 	    }
 	}
 	%c0 = %c1 unless $i == $V - 1;
@@ -3206,7 +3318,7 @@ sub _SPT_Bellman_Ford {
 	}
     }
 
-    return (\%p, \%d );
+    return (\%p, \%d);
 }
 
 sub SPT_Bellman_Ford {
@@ -3216,12 +3328,16 @@ sub SPT_Bellman_Ford {
 
     unless (defined $r) {
 	$r = $g->random_vertex();
+	return unless defined $r;
     }
 
     my $spt_bf = $g->get_graph_attribute('_spt_bf');
-    unless (defined $spt_bf && exists $spt_bf->{ $r } && $spt_bf->{ $r }->[ 0 ] == $g->[ _G ]) {
-	my ($p, $d) = $g->_SPT_Bellman_Ford($opt, $unseenh, $unseena, $r, $next, $code, $attr);
-	my $h = $g->new();
+    unless (defined $spt_bf &&
+	    exists $spt_bf->{ $r } && $spt_bf->{ $r }->[ 0 ] == $g->[ _G ]) {
+	my ($p, $d) =
+	    $g->_SPT_Bellman_Ford($opt, $unseenh, $unseena,
+				  $r, $next, $code, $attr);
+	my $h = $g->new;
 	for my $v (keys %$p) {
 	    my $u = $p->{ $v };
 	    $h->add_edge( $u, $v );
@@ -3234,7 +3350,11 @@ sub SPT_Bellman_Ford {
 	$g->set_graph_attribute('_spt_bf', $spt_bf);
     }
 
-    return $spt_bf->{ $r }->[ 1 ];
+    my $spt = $spt_bf->{ $r }->[ 1 ];
+
+    $spt->set_graph_attribute('SPT_Bellman_Ford_root', $r);
+
+    return $spt;
 }
 
 *SSSP_Bellman_Ford = \&SPT_Bellman_Ford;
@@ -3243,12 +3363,17 @@ sub SP_Bellman_Ford {
     my ($g, $u, $v) = @_;
     my $sptg = $g->SPT_Dijkstra(first_root => $u);
     my @path = ($v);
+    my %seen;
+    my $V = $g->vertices;
     my $p;
     while (defined($p = $sptg->get_vertex_attribute($v, 'p'))) {
+	last if exists $seen{$p};
 	push @path, $p;
 	$v = $p;
+	$seen{$p}++;
+	last if keys %seen == $V;
     }
-    @path = () if @path && $path[-1] ne $u;
+    # @path = () if @path && "$path[-1]" ne "$u";
     return reverse @path;
 }
 
@@ -3277,11 +3402,21 @@ sub APSP_Floyd_Warshall {
 sub transitive_closure_matrix {
     my $g = shift;
     my $tcm = $g->get_graph_attribute('_tcm');
+    if (defined $tcm) {
+	if (ref $tcm eq 'ARRAY') { # YECHHH!
+	    if ($tcm->[ 0 ] == $g->[ _G ]) {
+		$tcm = $tcm->[ 1 ];
+	    } else {
+		undef $tcm;
+	    }
+	}
+    }
     unless (defined $tcm) {
 	my $apsp = $g->APSP_Floyd_Warshall(@_);
 	$tcm = $apsp->get_graph_attribute('_tcm');
-	$g->set_graph_attribute('_tcm', $tcm);
+	$g->set_graph_attribute('_tcm', [ $g->[ _G ], $tcm ]);
     }
+
     return $tcm;
 }
 
@@ -3312,7 +3447,7 @@ sub is_reachable {
 sub for_shortest_paths {
     my $g = shift;
     my $c = shift;
-    my $t = Graph::TransitiveClosure->new($g, path => 1);
+    my $t = $g->transitive_closure_matrix;
     my @v = $g->vertices;
     my $n = 0;
     for my $u (@v) {
@@ -3328,14 +3463,15 @@ sub for_shortest_paths {
 sub _minmax_path {
     my $g = shift;
     my $min;
-    my $max = 0;
+    my $max;
     my $minp;
     my $maxp;
     $g->for_shortest_paths(sub {
 			       my ($t, $u, $v, $n) = @_;
 			       my $l = $t->path_length($u, $v);
+			       return unless defined $l;
 			       my $p;
-			       if ($l > $max) {
+			       if ($u ne $v && (!defined $max || $l > $max)) {
 				   $max = $l;
 				   $maxp = $p = [ $t->path_vertices($u, $v) ];
 			       }
@@ -3347,24 +3483,131 @@ sub _minmax_path {
     return ($min, $max, $minp, $maxp);
 }
 
-sub graph_diameter {
-    my $g = shift;
-    my ($min, $max, $minp, $maxp) = $g->_minmax_path(@_);
-    return $max;
-}
-
-sub longest_path {
+sub diameter {
     my $g = shift;
     my ($min, $max, $minp, $maxp) = $g->_minmax_path(@_);
     return defined $maxp ? (wantarray ? @$maxp : $max) : undef;
 }
 
+*graph_diameter = \&diameter;
+
+sub longest_path {
+    my ($g, $u, $v) = @_;
+    my $t = $g->transitive_closure_matrix;
+    if (defined $u) {
+	if (defined $v) {
+	    return wantarray ?
+		$t->path_vertices($u, $v) : $t->path_length($u, $v);
+	} else {
+	    my $max;
+	    my @max;
+	    for my $v ($g->vertices) {
+		next if $u eq $v;
+		my $l = $t->path_length($u, $v);
+		if (defined $l && (!defined $max || $l > $max)) {
+		    $max = $l;
+		    @max = $t->path_vertices($u, $v);
+		}
+	    }
+	    return wantarray ? @max : $max;
+	}
+    } else {
+	if (defined $v) {
+	    my $max;
+	    my @max;
+	    for my $u ($g->vertices) {
+		next if $u eq $v;
+		my $l = $t->path_length($u, $v);
+		if (defined $l && (!defined $max || $l > $max)) {
+		    $max = $l;
+		    @max = $t->path_vertices($u, $v);
+		}
+	    }
+	    return wantarray ? @max : @max - 1;
+	} else {
+	    my ($min, $max, $minp, $maxp) = $g->_minmax_path(@_);
+	    return defined $maxp ? (wantarray ? @$maxp : $max) : undef;
+	}
+    }
+}
+
+sub vertex_eccentricity {
+    my ($g, $u) = @_;
+    my $max;
+    for my $v ($g->vertices) {
+	next if $u eq $v;
+	my $l = $g->path_length($u, $v);
+	if (defined $l && (!defined $max || $l > $max)) {
+	    $max = $l;
+	}
+    }
+    return $max;
+}
+
 sub shortest_path {
+    my ($g, $u, $v) = @_;
+    my $t = $g->transitive_closure_matrix;
+    if (defined $u) {
+	if (defined $v) {
+	    return wantarray ?
+		$t->path_vertices($u, $v) : $t->path_length($u, $v);
+	} else {
+	    my $min;
+	    my @min;
+	    for my $v ($g->vertices) {
+		next if $u eq $v;
+		my $l = $t->path_length($u, $v);
+		if (defined $l && (!defined $min || $l < $min)) {
+		    $min = $l;
+		    @min = $t->path_vertices($u, $v);
+		}
+	    }
+	    return wantarray ? @min : $min;
+	}
+    } else {
+	if (defined $v) {
+	    my $min;
+	    my @min;
+	    for my $u ($g->vertices) {
+		next if $u eq $v;
+		my $l = $t->path_length($u, $v);
+		if (defined $l && (!defined $min || $l < $min)) {
+		    $min = $l;
+		    @min = $t->path_vertices($u, $v);
+		}
+	    }
+	    return wantarray ? @min : $min;
+	} else {
+	    my ($min, $max, $minp, $maxp) = $g->_minmax_path(@_);
+	    return defined $minp ? (wantarray ? @$minp : $min) : undef;
+	}
+    }
+}
+
+sub radius {
     my $g = shift;
     my ($min, $max, $minp, $maxp) = $g->_minmax_path(@_);
     return defined $minp ? (wantarray ? @$minp : $min) : undef;
 }
 
+sub center_vertices {
+    my ($g, $delta) = @_;
+    $delta = 0 unless defined $delta;
+    $delta = abs($delta);
+    my @c;
+    my $r = $g->radius;
+    if (defined $r) {
+	for my $v ($g->vertices) {
+	    my $e = $g->vertex_eccentricity($v);
+	    next unless defined $e;
+	    push @c, $v if abs($e - $r) <= $delta;
+	}
+    }
+    return @c;
+}
+
+*centre_vertices = \&center_vertices;
+
 sub average_path_length {
     my $g = shift;
     my @A = @_;
